/*
 * fonts.c
 * Copyright (C) 1998,1999 A.J. van Os
 *
 * Description:
 * Functions to deal with fonts
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "antiword.h"

static font		tFontCurr = -1;
/* Font Translation Table */
static int		iFontTableRecords = 0;
static font_table_type	*pFontTable = NULL;

/*
 * Find the given font in the font table
 *
 * returns the index into the FontTable, -1 if not found
 */
static int
iFindAcornFont(unsigned char ucWordFontnumber, unsigned char ucFontstyle)
{
	int	iIndex;

	for (iIndex = 0; iIndex < iFontTableRecords; iIndex++) {
		if (pFontTable[iIndex].ucWordFontnumber ==
							ucWordFontnumber &&
		    pFontTable[iIndex].ucFontstyle == ucFontstyle) {
			return iIndex;
		}
	}
	return -1;
} /* end of iFindAcornFont */

/*
 * pOpenFontTableFile - open the Font translation file
 * Copy the file to the proper place if necessary.
 *
 * Returns the file pointer or NULL
 */
static FILE *
pOpenFontTableFile(void)
{
	FILE	*pFileR, *pFileW;
	char	*szFontNamesFile;
	size_t	tSize;
	BOOL	bFailed;
	char	acBuffer[256];

	pFileR = fopen("<AntiWord$FontNamesFile>", "r");
	if (pFileR != NULL) {
		/* The font table is already in the right directory */
		return pFileR;
	}

	szFontNamesFile = getenv("AntiWord$FontNamesSave");
	if (szFontNamesFile == NULL) {
		werr(0, "Warning: Name of the FontNames file not found");
		return NULL;
	}
	DBG_MSG(szFontNamesFile);

	pFileR = fopen("<AntiWord$Dir>.Resources.Default", "r");
	if (pFileR == NULL) {
		werr(0, "I can't find 'Resources.Default'");
		return NULL;
	}
	/* Here the default font translation table is known to exist */

	if (!bMakeDirectory(szFontNamesFile)) {
		werr(0,
		"I can't make a directory for the FontNames file");
		return NULL;
	}
	/* Here the proper directory is known to exist */

	pFileW = fopen(szFontNamesFile, "w");
	if (pFileW == NULL) {
		(void)fclose(pFileR);
		werr(0, "I can't create a default FontNames file");
		return NULL;
	}
	/* Here the proper directory is known to be writeable */

	/* Copy the default FontNames file */
	bFailed = FALSE;
	while (!feof(pFileR)) {
		tSize = fread(acBuffer, 1, sizeof(acBuffer), pFileR);
		if (ferror(pFileR)) {
			DBG_MSG("Read error");
			bFailed = TRUE;
			break;
		}
		if (fwrite(acBuffer, 1, tSize, pFileW) != tSize) {
			DBG_MSG("Write error");
			bFailed = TRUE;
			break;
		}
	}
	(void)fclose(pFileW);
	(void)fclose(pFileR);
	if (bFailed) {
		DBG_MSG("Copying the FontNames file failed");
		(void)remove(szFontNamesFile);
		return NULL;
	}
	return fopen(szFontNamesFile, "r");
} /* end of pOpenFontTableFile */

/*
 * See if the fontname from the Word file matches the fontname from the
 * font translation file.
 * If iBytesPerChar is one than szWord is in ISO-8859-1 (Word 6/7),
 * if iBytesPerChar is two than szWord is in Unicode (Word 8/97).
 */
static BOOL
bFontEqual(const char *szWord, const char *szTable, int iBytesPerChar)
{
	const char	*pcTmp1, *pcTmp2;

	fail(szWord == NULL || szTable == NULL);
	fail(iBytesPerChar != 1 && iBytesPerChar != 2);

	for (pcTmp1 = szWord, pcTmp2 = szTable;
	     *pcTmp1 != '\0';
	     pcTmp1 += iBytesPerChar, pcTmp2++) {
		if (tolower(*pcTmp1) != tolower(*pcTmp2)) {
			return FALSE;
		}
	}
	return *pcTmp2 == '\0';
} /* end of bFontEqual */

/*
 * vCreateFontTable - Create and initialize the internal font table
 */
static void
vCreateFontTable(void)
{
	font_table_type	*pTmp;
	int	iSize, iNbr;

	if (iFontTableRecords <= 0) {
		pFontTable = NULL;
		iFontTableRecords = 0;
		return;
	}

	/* Create the font table */
	iSize = iFontTableRecords * sizeof(*pFontTable);
	pFontTable = xmalloc(iSize);
	/* Initialize the font table */
	(void)memset(pFontTable, 0, iSize);
	for (iNbr = 0, pTmp = pFontTable;
	     pTmp < pFontTable + iFontTableRecords;
	     iNbr++, pTmp++) {
		pTmp->ucWordFontnumber = iNbr / 4;
		switch (iNbr % 4) {
		case 0:
			pTmp->ucFontstyle = FONT_REGULAR;
			break;
		case 1:
			pTmp->ucFontstyle = FONT_BOLD;
			break;
		case 2:
			pTmp->ucFontstyle = FONT_ITALIC;
			break;
		case 3:
			pTmp->ucFontstyle = FONT_ITALIC|FONT_BOLD;
			break;
		default:
			DBG_DEC(iNbr);
			break;
		}
	}
} /* end of vCreateFontTable */

/*
 * vMinimizeFontTable - make the font table as small as possible
 */
static void
vMinimizeFontTable(void)
{
	const font_block_type	*pFont;
	font_table_type		*pTmp;
	int	iUnUsed;
	BOOL	bMustAddTableFont;

	NO_DBG_MSG("vMinimizeFontTable");

	if (pFontTable == NULL || iFontTableRecords <= 0) {
		return;
	}

#if 0
	DBG_MSG("Before");
	DBG_DEC(iFontTableRecords);
	for (pTmp = pFontTable;
	     pTmp < pFontTable + iFontTableRecords;
	     pTmp++) {
		DBG_DEC(pTmp->ucWordFontnumber);
		DBG_HEX(pTmp->ucFontstyle);
		DBG_MSG(pTmp->szWordFontname);
		DBG_MSG(pTmp->szAcornFontname);
	}
#endif /* DEBUG */

	/* Default font/style is by definition in use */
	pFontTable[0].ucInUse = TRUE;

	/* See which fonts/styles are really being used */
	bMustAddTableFont = TRUE;
	pFont = NULL;
	while((pFont = pGetNextFontInfoListItem(pFont)) != NULL) {
		pTmp = pFontTable + 4 * (int)pFont->ucFontnumber;
		if (bIsBold(pFont->ucFontstyle)) {
			pTmp++;
		}
		if (bIsItalic(pFont->ucFontstyle)) {
			pTmp += 2;
		}
		if (pTmp >= pFontTable + iFontTableRecords) {
			continue;
		}
		if (StrEq(pTmp->szAcornFontname, TABLE_FONT)) {
			/* The table font is already present */
			bMustAddTableFont = FALSE;
		}
		pTmp->ucInUse = TRUE;
	}

	/* Remove the unused font entries from the font table */
	iUnUsed = 0;
	for (pTmp = pFontTable;
	     pTmp < pFontTable + iFontTableRecords;
	     pTmp++) {
		if (!pTmp->ucInUse) {
			iUnUsed++;
			continue;
		}
		if (iUnUsed > 0) {
			fail(pTmp - iUnUsed <= pFontTable);
			*(pTmp - iUnUsed) = *pTmp;
		}
	}
	iFontTableRecords -= iUnUsed;
	if (bMustAddTableFont) {
		pTmp = pFontTable + iFontTableRecords;
	  	pTmp->ucWordFontnumber = (pTmp - 1)->ucWordFontnumber + 1;
		pTmp->ucFontstyle = FONT_REGULAR;
		pTmp->ucInUse = TRUE;
		strcpy(pTmp->szWordFontname, "Extra Table Font");
		strcpy(pTmp->szAcornFontname, TABLE_FONT);
		iFontTableRecords++;
		iUnUsed--;
	}
	if (iUnUsed > 0) {
		/* Resize the font table */
		pFontTable = xrealloc(pFontTable,
				iFontTableRecords * sizeof(*pFontTable));
	}
#if 0
	/* Report the used but untranslated fontnames */
	for (pTmp = pFontTable;
	     pTmp < pFontTable + iFontTableRecords;
	     pTmp++) {
		if (pTmp->szAcornFontname[0] == '\0') {
			DBG_MSG(pTmp->szWordFontname);
			werr(0, "%s is not in the FontNames file",
				pTmp->szWordFontname);
		}
	}
#endif /* 0 */
#if 0
	DBG_MSG("After");
	DBG_DEC(iFontTableRecords);
	for (pTmp = pFontTable;
	     pTmp < pFontTable + iFontTableRecords;
	     pTmp++) {
		DBG_DEC(pTmp->ucWordFontnumber);
		DBG_HEX(pTmp->ucFontstyle);
		DBG_MSG(pTmp->szWordFontname);
		DBG_MSG(pTmp->szAcornFontname);
	}
#endif /* DEBUG */
} /* end of vMinimizeFontTable */

/*
 * bReadFontFile - read and check a line from the font translation file
 *
 * returns TRUE when a correct line has been read, otherwise FALSE
 */
static BOOL
bReadFontFile(FILE *pFontTableFile, char *szWordFont,
	int *piItalic, int *piBold, char *szAcornFont, int *piSpecial)
{
	char	*pcTmp;
	int	iFields;
	char	szLine[81];

	fail(szWordFont == NULL || szAcornFont == NULL);
	fail(piItalic == NULL || piBold == NULL || piSpecial == NULL);

	while (fgets(szLine, sizeof(szLine), pFontTableFile) != NULL) {
		if (szLine[0] == OPTIONS_COMMENT_CHAR ||
		    szLine[0] == '\n' ||
		    szLine[0] == '\r') {
			continue;
		}
		iFields = sscanf(szLine, "%1s%[^,],%d,%d,%1s%[^,],%d",
		    	&szWordFont[0], &szWordFont[1], piItalic, piBold,
		    	&szAcornFont[0], &szAcornFont[1], piSpecial);
		if (iFields != 7) {
			pcTmp = strchr(szLine, '\r');
			if (pcTmp != NULL) {
				*pcTmp = '\0';
			}
			pcTmp = strchr(szLine, '\n');
			if (pcTmp != NULL) {
				*pcTmp = '\0';
			}
			DBG_DEC(iFields);
			werr(0, "Syntax error in: '%s'", szLine);
			continue;
		}
		if (strlen(szWordFont) >=
		    		sizeof(pFontTable[0].szWordFontname) ||
		    strlen(szAcornFont) >=
		    		sizeof(pFontTable[0].szAcornFontname)) {
			continue;
		}
		/* The current line passed all the tests */
		return TRUE;
	}
	return FALSE;
} /* end of bReadFontFile */

/*
 * vCreate6FontTable - create a font table for Word 6/7
 */
void
vCreate6FontTable(FILE *pFile, int iStartBlock,
	const int *aiBBD, int iBBDLen,
	const unsigned char *aucHeader)
{
	FILE	*pFontTableFile;
	font_table_type	*pTmp;
	char	*szFont, *szAltFont;
	unsigned char	*aucBuffer;
	int	iNbr, iPos, iRecLen, iOffsetAltName;
	int	iBold, iItalic, iSpecial, iBeginFontInfo, iFontInfoLen;
	char	szWordFont[81], szAcornFont[81];

	fail(pFile == NULL || aucHeader == NULL);
	fail(iStartBlock < 0);
	fail(aiBBD == NULL || iBBDLen < 0);

	pFontTableFile = pOpenFontTableFile();
	if (pFontTableFile == NULL) {
		/* No translation table file, no translation table */
		return;
	}

	iBeginFontInfo = (int)ulGetLong(0xd0, aucHeader);
	DBG_HEX(iBeginFontInfo);
	iFontInfoLen = (int)ulGetLong(0xd4, aucHeader);
	DBG_DEC(iFontInfoLen);
	fail(iFontInfoLen < 9);

	aucBuffer = xmalloc(iFontInfoLen);
	if (!bReadBuffer(pFile, iStartBlock,
			aiBBD, iBBDLen, BIG_BLOCK_SIZE,
			aucBuffer, iBeginFontInfo, iFontInfoLen)) {
		xfree(aucBuffer);
		(void)fclose(pFontTableFile);
		return;
	}
	DBG_DEC(usGetWord(0, aucBuffer));

	/* Compute the maximum number of entries in the font table */
	iFontTableRecords = 0;
	iPos = 2;
	while (iPos + 6 < iFontInfoLen) {
		iRecLen = (int)ucGetByte(iPos, aucBuffer);
		NO_DBG_DEC(iRecLen);
		iOffsetAltName = (int)ucGetByte(iPos + 5, aucBuffer);
		NO_DBG_MSG(aucBuffer + iPos + 6);
		NO_DBG_MSG_C(iOffsetAltName > 0,
				aucBuffer + iPos + 6 + iOffsetAltName);
		iPos += iRecLen + 1;
		iFontTableRecords++;
	}
	iFontTableRecords *= 4;	/* Regular, Bold, Italic and Bold/italic */
	iFontTableRecords++;	/* One extra for the table-font */
	vCreateFontTable();

	/* Read the font translation file */
	while (bReadFontFile(pFontTableFile, szWordFont,
			&iItalic, &iBold, szAcornFont, &iSpecial)) {
		pTmp = pFontTable;
		if (iBold != 0) pTmp++;
		if (iItalic != 0) pTmp += 2;
		iNbr = 0;
		iPos = 2;
		while (iPos + 6 < iFontInfoLen) {
			iRecLen = (int)ucGetByte(iPos, aucBuffer);
			iOffsetAltName = (int)ucGetByte(iPos + 5, aucBuffer);
			szFont = (char *)aucBuffer + iPos + 6;
			szAltFont = szFont + iOffsetAltName;
			if (bFontEqual(szFont, szWordFont, 1) ||
			    iOffsetAltName > 0 &&
			     bFontEqual(szAltFont, szWordFont, 1) ||
			    pTmp->szWordFontname[0] == '\0' &&
			     szWordFont[0] == '*' &&
			     szWordFont[1] == '\0') {
			  	strcpy(pTmp->szWordFontname, szFont);
				strcpy(pTmp->szAcornFontname, szAcornFont);
			}
			pTmp += 4;
			iNbr++;
			iPos += iRecLen + 1;
		}
	}
	vMinimizeFontTable();
	xfree(aucBuffer);
	(void)fclose(pFontTableFile);
} /* end of vCreate6FontTable */

void
vCreate8FontTable(FILE *pFile, const pps_info_type *pPPS,
	const int *aiBBD, int iBBDLen, const int *aiSBD, int iSBDLen,
	const unsigned char *aucHeader)
{
	FILE	*pFontTableFile;
	font_table_type	*pTmp;
	const int	*aiBlockDepot;
	char		*szFont, *szAltFont;
	unsigned char	*aucBuffer;
	int	iTableStartBlock, iTableSize, iBlockDepotLen, iBlockSize;
	int	iNbr, iPos, iRecLen, iOffsetAltName;
	int	iBold, iItalic, iSpecial, iBeginFontInfo, iFontInfoLen;
	unsigned short	usDocStatus;
	char	szWordFont[81], szAcornFont[81];

	fail(pFile == NULL || pPPS == NULL || aucHeader == NULL);
	fail(aiBBD == NULL || iBBDLen < 0);

	pFontTableFile = pOpenFontTableFile();
	if (pFontTableFile == NULL) {
		/* No translation table file, no translation table */
		return;
	}

	iBeginFontInfo = (int)ulGetLong(0x112, aucHeader);
	DBG_HEX(iBeginFontInfo);
	iFontInfoLen = (int)ulGetLong(0x116, aucHeader);
	DBG_DEC(iFontInfoLen);
	fail(iFontInfoLen < 46);

	/* Use 0Table or 1Table? */
	usDocStatus = usGetWord(0x0a, aucHeader);
	if (usDocStatus & BIT(9)) {
		iTableStartBlock = pPPS->t1Table.iSb;
		iTableSize = pPPS->t1Table.iSize;
	} else {
		iTableStartBlock = pPPS->t0Table.iSb;
		iTableSize = pPPS->t0Table.iSize;
	}
	DBG_DEC(iTableStartBlock);
	if (iTableStartBlock < 0) {
		DBG_DEC(iTableStartBlock);
		DBG_MSG("No fontname table");
		return;
	}
	DBG_HEX(iTableSize);
	if (iTableSize < MIN_SIZE_FOR_BBD_USE) {
	  	/* Use the Small Block Depot */
		aiBlockDepot = aiSBD;
		iBlockDepotLen = iSBDLen;
		iBlockSize = SMALL_BLOCK_SIZE;
	} else {
	  	/* Use the Big Block Depot */
		aiBlockDepot = aiBBD;
		iBlockDepotLen = iBBDLen;
		iBlockSize = BIG_BLOCK_SIZE;
	}
	aucBuffer = xmalloc(iFontInfoLen);
	if (!bReadBuffer(pFile, iTableStartBlock,
			aiBlockDepot, iBlockDepotLen, iBlockSize,
			aucBuffer, iBeginFontInfo, iFontInfoLen)) {
		xfree(aucBuffer);
		(void)fclose(pFontTableFile);
		return;
	}
	NO_DBG_PRINT_BLOCK(aucBuffer, iFontInfoLen);

	/* Get the maximum number of entries in the font table */
	iFontTableRecords = (int)usGetWord(0, aucBuffer);
	iFontTableRecords *= 4;	/* Regular, Bold, Italic and Bold/italic */
	iFontTableRecords++;	/* One extra for the table-font */
	vCreateFontTable();

	/* Read the font translation file */
	while (bReadFontFile(pFontTableFile, szWordFont,
			&iItalic, &iBold, szAcornFont, &iSpecial)) {
		pTmp = pFontTable;
		if (iBold != 0) pTmp++;
		if (iItalic != 0) pTmp += 2;
		iNbr = 0;
		iPos = 4;
		while (iPos + 40 < iFontInfoLen) {
			iRecLen = (int)ucGetByte(iPos, aucBuffer);
#if 0
			iOffsetAltName = (int)ucGetByte(iPos + 5, aucBuffer);
#else
			iOffsetAltName = 0;
#endif /* 0 */
			szFont = (char *)aucBuffer + iPos + 40;
			szAltFont = szFont + iOffsetAltName;
			if (bFontEqual(szFont, szWordFont, 2) ||
			    iOffsetAltName > 0 &&
			     bFontEqual(szAltFont, szWordFont, 2) ||
			    pTmp->szWordFontname[0] == '\0' &&
			     szWordFont[0] == '*' &&
			     szWordFont[1] == '\0') {
				unicpy(pTmp->szWordFontname, szFont);
				strcpy(pTmp->szAcornFontname, szAcornFont);
			}
			pTmp += 4;
			iNbr++;
			iPos += iRecLen + 1;
		}
	}
	vMinimizeFontTable();
	xfree(aucBuffer);
	(void)fclose(pFontTableFile);
} /* end of vCreate8FontTable */

void
vDestroyFontTable(void)
{
	DBG_MSG("vDestroyFontTable");

	if (pFontTable != NULL) {
		xfree(pFontTable);
		pFontTable = NULL;
	}
	iFontTableRecords = 0;
} /* end of vDestroyFontTable */

/*
 * vCloseFont - close the current font, if there is any
 */
void
vCloseFont()
{
	os_error	*e;

	DBG_MSG("vCloseFont");

	if (tFontCurr == (font)-1) {
		return;
	}
	e = font_lose(tFontCurr);
	if (e != NULL) {
		werr(0, "Close font error %d: %s", e->errnum, e->errmess);
	}
	tFontCurr = -1;
} /* end of vCloseFont */

/*
 * tOpenFont - make the given font the current font
 *
 * Returns the font reference number for use in a draw file
 */
draw_fontref
tOpenFont(unsigned char ucWordFontNumber,
		unsigned char ucFontstyle, unsigned char ucWordFontsize)
{
	os_error	*e;
	font	tFont;
	int	iIndex;

	DBG_MSG("tOpenFont");
	DBG_DEC(ucWordFontNumber);
	DBG_DEC(ucWordFontsize);

	iIndex = iFindAcornFont(ucWordFontNumber, ucFontstyle);
	if (iIndex < 0 || iIndex >= 255) {
		DBG_DEC(iIndex);
		tFontCurr = (font)-1;
		return 0;
	}
	if (pFontTable[iIndex].szAcornFontname[0] == '\0') {
		DBG_MSG(pFontTable[iIndex].szWordFontname);
		tFontCurr = (font)-1;
		return 0;
	}
	DBG_MSG(pFontTable[iIndex].szAcornFontname);
	e = font_find(pFontTable[iIndex].szAcornFontname,
			(int)ucWordFontsize * 8, (int)ucWordFontsize * 8,
			0, 0, &tFont);
	if (e != NULL) {
		switch (e->errnum) {
		case 523:
			werr(0, "%s", e->errmess);
			break;
		default:
			werr(0, "Open font error %d: %s",
				e->errnum, e->errmess);
			break;
		}
		tFontCurr = (font)-1;
		return 0;
	}
	tFontCurr = tFont;
	DBG_DEC(tFontCurr);
	return iIndex + 1;
} /* end of tOpenFont */

/*
 * tOpenTableFont - make the table font the current font
 *
 * Returns the font reference number for use in a draw file
 */
draw_fontref
tOpenTableFont(unsigned char ucWordFontsize)
{
	int	iIndex;

	DBG_MSG("tOpenTableFont");

	/* Search for the font in the table */
	for (iIndex = 0; iIndex < iFontTableRecords; iIndex++) {
		if (pFontTable[iIndex].ucFontstyle == FONT_REGULAR &&
		    StrEq(pFontTable[iIndex].szAcornFontname, TABLE_FONT)) {
			return tOpenFont(pFontTable[iIndex].ucWordFontnumber,
					FONT_REGULAR, ucWordFontsize);
		}
	}
	tFontCurr = (font)-1;
	return 0;
} /* end of tOpenTableFont */

/*
 * iComputeStringWidth - compute the string width
 *
 * Returns string width in millipoints
 */
int
iComputeStringWidth(char *szString, int iStringLength)
{
	font_string	tStr;
	os_error	*e;

	fail(szString == NULL || iStringLength < 0);

	if (szString[0] == '\0' || iStringLength <= 0) {
		/* Empty string */
		return 0;
	}
	if (iStringLength == 1 && szString[0] == TABLE_SEPARATOR) {
		/* Font_strwidth doesn't like control characters */
		return 0;
	}
	if (tFontCurr == (font)-1) {
		/* No current font, use systemfont */
		return iChar2MilliPoints(iStringLength);
	}
	tStr.s = szString;
	tStr.x = INT_MAX;
	tStr.y = INT_MAX;
	tStr.split = -1;
	tStr.term = iStringLength;
	e = font_strwidth(&tStr);
	if (e == NULL) {
		return tStr.x;
	}
	DBG_DEC(e->errnum);
	DBG_MSG(e->errmess);
	DBG_DEC(iStringLength);
	DBG_MSG(szString);
	werr(0, "String width error %d: %s", e->errnum, e->errmess);
	return iChar2MilliPoints(iStringLength);
} /* end of iComputeStringWidth */

/*
 * vGetFontTranslationTable - make the table known to the outside world
 */
void
vGetFontTranslationTable(font_table_type **ppTable, int *piRecords)
{
	if (ppTable != NULL) {
		*ppTable = pFontTable;
	}
	if (piRecords != NULL) {
		*piRecords = iFontTableRecords;
	}
} /* end of vGetFontTranslationTable */

/*
 * Translate a Word colournumber into a true colour for use in a drawfile
 *
 * Returns the true colour
 */
int
iColour2Colour(unsigned char ucWordColour)
{
	static const int	aiColourTable[] = {
		/*  0 */	0x00000000,
		/*  1 */	0x00000000,
		/*  2 */	0xff000000,
		/*  3 */	0xffff0000,
		/*  4 */	0x00ff0000,
		/*  5 */	0xff00ff00,
		/*  6 */	0x0000ff00,
		/*  7 */	0x00ffff00,
		/*  8 */	0xffffff00,
		/*  9 */	0x80000000,
		/* 10 */	0x80800000,
		/* 11 */	0x00800000,
		/* 12 */	0x80008000,
		/* 13 */	0x00000800,
		/* 14 */	0x00808000,
		/* 15 */	0x80808000,
		/* 16 */	0xc0c0c000,
	};
	if (ucWordColour >= elementsof(aiColourTable)) {
		return aiColourTable[0];
	}
	return aiColourTable[ucWordColour];
} /* end of iColour2Colour */
